! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: rgsoffsetcal_mod
!
! Authors:  Giuseppe Vacanti, European Space Agency, 1999
!          
! Description: calculates the mean and rms of the PHA distribution of
! an RGS diagnostic image. 
! 
! Comments: 
!   TODO --
!       - more diagnostic messages
!       - use OAL
!       - use DAL compound notation dataset:block
!       - add printing of iterations to gaussian_residuals
!       - switch NPRINT=1 depending on verbosity level
!       - make bin size a parameter
!       - encapsulate fitting
!       - allow for more functions to fit
!       - issue warning if BLOCKNO indicates dataset is only one part
!         of the full image.
! 
!- 

module rgsoffsetcalc_mod

  use dal
  use oal
  use param  
  use saslib
  use tablesort_mod
  use errorhandling
  use histogram
  use imagelib
  use CalOalDefs
!!!  use eventAttributes
!!!  use odfNames

  implicit none 
  private 

  type(histogram_t), pointer :: histo
  real :: nblanks_fraction = 0.1 !

  public rgsoffsetcalc

contains

  subroutine rgsoffsetcalc    

    type(DataSetT) :: set, hset, outset
    type(ArrayT)   :: arr
    type(BlockT)   :: blk

    type(TableT) :: tbl
    type(ColumnT) :: col

    integer(kind=int16), dimension(:,:), pointer :: image
    integer(kind=int32) :: nbins, i, j, lo, hi, node, ccdnode, inode
    integer(kind=int32), dimension(1:4) :: nodes = (/ 1, 1, 2 , 1/)
    real(kind=single) :: min, max, pha, euclidenorm
    real(single), pointer :: histovals(:)
    character(len=128) :: premsg
    real(kind=single), allocatable, dimension(:) :: minima, maxima
    character(len=5) :: calscope = ""
    character(len=5) :: blockno
    character(len=128) :: blkname, label
    logical :: ignorecalscope = .false.

    ! Upper case variables are needed for SLATEC
    integer(kind=int32), parameter :: N = 3
    real(kind=single), dimension(N) :: X
    integer(kind=int32) :: M
    real(kind=single), allocatable, dimension(:) :: FVEC
    real(kind=single) :: TOL = 1e-3
    integer(kind=int32), parameter :: IOPT = 1 ! calculate Jacobian by forward differencing
    integer(kind=int32) :: NPRINT = 0
    integer(kind=int32) :: INFO
    integer(kind=int32), dimension(N) :: IW
    integer(kind=int32) :: LWA
    integer(kind=int32), dimension(:), allocatable :: WA

    real(kind=single) ENORM
    external ENORM

    ! Messages from SLATEC (copied from documentation)
    character(len=128), dimension(0:7) :: infomsg = (/ &
         & "improper input parameters.                                                                                      ", &
         & "algorithm estimates that the relative error in the sum of squares is at most TOL.                               ", &
         & "algorithm estimates that the relative error between X and the solution is at most TOL.                          ", &
         & "algorithm estimates that the relative error in the sum of squares and between X and the solution is at most TOL.", &
         & "FVEC is orthogonal to the columns of the Jacobian to machine precision.                                         ", &
         & "number of calls to FCN has reached 100*(N+1) for IOPT=2 or 3 or 200*(N+1) for IOPT=1.                           ", &
         & "TOL is too small.  No further reduction in the sum of squares is possible.                                      ", &
         & "TOL is too small.  No further improvement in the approximate solution X is possible.                            " &
         & /)

    ! -----------------------------------------------------------------------------

    if(booleanParameter("withhistogramset"))then
       hset = dataSet(stringParameter("histogramset"), CREATE)
    end if

    ! Identify the calscope (INSTRUME attribute). Take the first one found.
    ! Sets with a different scope will be ignored.
    ignorecalscope = .true.
    do j=1,parameterCount("sets")
       set = dataSet(stringParameter("sets",j-1), READ)
       blkname = trim(stringParameter("sets",j-1)) // ":"
       blk = block(set, blkname, READ)
       if(hasAttribute(blk,"INSTRUME"))then
          calscope = stringAttribute(blk,"INSTRUME")
          call release(blk)
          call release(set)
          ignorecalscope = .false.
          exit
       end if
       call release(blk)
       call release(set)
    end do
    if(ignorecalscope)then
       write(errstr,*)"Cannot identify a valid calscope (INSTRUME attribute). " // trim(stringParameter("outset")) // &
            " will not be a valid CCF constituent."       
       call warning('CalibrationScopeIgnored')
    else
       write(errstr,*)"Identified the following calibration scope = " // calscope // &
            ". Any other calibration scope will be ignored."       
       call message(appMsg, Silent)
    endif
    outset = dataSet(stringParameter("outset"), CREATE)
    call initCcfSet(outset, calscope)

    !
    !-----------------------------------------------------------------
    !
    ! This is to be done only if --sets contains at least one value,
    ! or some of the Fortran statements will fail.
    if(parameterCount("sets")>0)then

       ! Calculate common min and max. This is required so that if
       ! the use wants to store the spectra they all have the same
       ! length and can be easily accomodated in the same table.
       allocate(minima(parameterCount("sets")), maxima(parameterCount("sets")))
       do j=1, parameterCount("sets")
          set = dataSet(stringParameter("sets",j-1), READ)
          arr = getDiagnosticArray(set)
          image => int16Array2Data(arr)
          minima(j) = minval(image)
          maxima(j) = maxval(image)
          call release(arr)
          call release(set)
       enddo
       min = minval(minima)
       max = maxval(maxima) + 1
       nbins = max - min + 1

       ! This test would make more sense on a per dataset basis. It
       ! is possible that only one dataset has enough non zero
       ! channels, and the others have no useful channels.
       if(nbins < N)then
          write(errstr,*)trim(premsg) // 'not enough data bins'
          call error('nothingToFit')
       end if
       do j = 0, parameterCount("sets") - 1
          set = dataSet(stringParameter("sets",j), READ)
          premsg = '(' // trim(name(set)) // ')'
          if(hasAttribute(set,"INSTRUME") .and. .not. ignorecalscope)then
             if(trim(stringAttribute(set,"INSTRUME")) /= calscope)then
                write(errstr,*)trim(premsg) // " Mismatch in calibration scope. Processing " // calscope // " and found " // &
                     trim(stringAttribute(set,"INSTRUME"))
                call warning("DatasetIgnored")
                call release(set)
                cycle
             end if
          end if
          if(.not. validateImage(set))then
             write(errstr,*)trim(premsg) // " Dataset unlikely to yield a meaningful fit"
             call warning("DatasetIgnored")
             call release(set)
             cycle
          end if
          arr = getDiagnosticArray(set)
          image => int16Array2Data(arr)          

          ! For genuine RGS files CCDNODE encodes how many read out
          ! nodes were used. See ODF ICD.
          !    CCDNODE = 1 => read out through node 0
          !    CCDNODE = 2 => read out through node 1
          !    CCDNODE = 3 => read out through nodes 0 and 1
          if(hasAttribute(arr,'CCDNODE'))then
             ccdnode = int32Attribute(attribute(arr,'CCDNODE'))
             write(errstr,*)trim(premsg) // ' Found attribute CCDNODE = ', ccdnode
          else
             ccdnode = 1
             write(errstr,*)trim(premsg) // ' Did not find attribute CCDNODE. Using the whole image as if it came from node 0.'
          end if
          call message(appMsg, Sparse)

          if(hasAttribute(arr,'BLOCKNO'))then
             blockno = stringAttribute(attribute(arr,'BLOCKNO'))
             if(trim(blockno) /= 'X')then
                write(errstr,*)trim(premsg) // ' BLOCKNO is not X but ' // blockno // &
                     '. These diagnostic images should have been combined into one.'
                call warning("DiagnosticImagesNotCombined")
             endif
          end if
          
          select case(ccdnode)
          case(1,3)
             inode=0
          case(2)
             inode=1
!          case(4)
!             inode=2 ! This will be translated to NULL
          end select

          do node = 0, nodes(ccdnode) - 1
             lo = node * size(image,1)/nodes(ccdnode)
             hi = lo + size(image,1)/nodes(ccdnode) - 1
             write(label,*)trim(name(set)),' lo = ',lo,' hi = ',hi
             call initHisto()
             histo => newHisto(name = "pedestal", &
                  & nbins = nbins, &
                  & min = min, max = max, &
                  & title = "pulse height spectrum", label = label)
             call addHisto(histo, real(image(lo:hi,:)))
             histovals => getArray(histo)

             ! Estimates for mean/standard deviation/normalization
             X(1) = 0.
             X(2) = 0.
             do i=0,size(histovals)-1
                pha = getMinimum(histo) + (i + 0.5) * getBinsize(histo)
                X(1) = X(1) + histovals(i) * pha
                X(2) = X(2) + histovals(i) * pha**2
             end do
             X(1) = X(1)/sum(histovals)                ! mean
             X(2) = sqrt(X(2)/sum(histovals)-X(1)**2)  ! sigma
             X(3) = maxval(histovals)                  ! normalization
             write(errstr,*)trim(premsg) // ' estimated parameters (m,s,n)',X
             call message(appMsg,Noisy)
             ! Set up and go
             M = getNbins(histo)
             allocate(FVEC(0:M-1))
             LWA = N*(M+5)+M
             allocate(WA(LWA))
             call SNLS1E(gaussian_residuals, IOPT, M, N, X, FVEC, TOL, NPRINT,&
                  & INFO, IW, WA, LWA)
             write(errstr,*)trim(premsg) // " " // trim(infomsg(INFO))
             call message(appMsg, Verbose)
             euclidenorm = ENORM(M,FVEC)
             deallocate(FVEC)
             deallocate(WA)
             write(errstr,*)trim(premsg) // ' fitted parameters (m,s,n)',X
             call message(appMsg,Verbose)

             call storeFit(arr,inode,X,euclidenorm,outset)

             if(booleanParameter("withhistogramset"))then
                call storeHisto(histo,hset)
             end if

             ! If we're going through the loop 2, we do CCD node 0 first,
             ! then 1.
             inode = inode + 1 

          end do
          call freeHisto()
          call release(arr)
          call release(set)

       end do
       if(booleanParameter("withhistogramset"))then
          call release(hset)
       end if
    end if
    tbl = table(outset, "RGSOFFSETS")
    col = column(tbl, "TIMESORT")
    call tableSort(tbl, col, 1)
    call release(col)
    call release(tbl)
    call release(outset)
  end subroutine rgsoffsetcalc

  subroutine storeFit(arr,inode,X,euclidenorm,set)
    use dal
    type(DataSetT), intent(in) :: set
    type(ArrayT), intent(in) :: arr
    real(kind=single), dimension(:), intent(in) :: X
    real(kind=single), intent(in) :: euclidenorm
    integer(kind=int32), intent(in) :: inode

    real(kind=double), dimension(:), pointer :: tx
    integer(kind=int32), dimension(:), pointer :: sx
    integer(kind=int8), dimension(:), pointer :: ix
    real(kind=single), dimension(:), pointer :: fx
    type(TableT) :: tbl
    type(ColumnT) :: c_time, c_ccd, c_ccdocb, c_mean, c_stddev, c_norm, c_node, c_euclidenorm, c_sort
    integer(kind=int32) :: j, ftfine, ftcoarse
    real(kind = double), dimension(:), pointer :: time => null()
    real(kind = double), dimension(0:0) :: obt
    

    tbl = table(set, "RGSOFFSETS")
    c_time = column(tbl, "TIME", MODIFY)
    c_sort = column(tbl, "TIMESORT", MODIFY)
    c_ccd = column(tbl, "CCD_ID", MODIFY)
    c_node = column(tbl, "NODE_ID", MODIFY)
    c_ccdocb = column(tbl, "CCD_OCB", MODIFY)
    c_mean = column(tbl, "MEAN", MODIFY)
    c_stddev = column(tbl, "STD_DEV", MODIFY)
    c_norm = column(tbl, "NORMALIZATION", MODIFY)
    c_euclidenorm = column(tbl, "EUCLIDIAN_NORM", MODIFY)

    call insertRows(tbl, numberOfRows(tbl), 1)
    j = numberOfRows(tbl) - 1
    tx => real64Data(c_time)
    sx => int32Data(c_sort)
    ! Make several attempts to get the correct time for this exposure
    if(hasAttribute(arr, "FTFINE") .and. hasAttribute(arr, "FTCOARSE"))then
       ftfine = int32Attribute(arr, "FTFINE");
       ftcoarse = int32Attribute(arr, "FTCOARSE");
       write(errstr,*)'Found FTFINE/FTCOARSE: ', ftfine, ftcoarse
       call message(AppMsg, Verbose)
       obt(0) = ftcoarse + ftfine/65536. ! ftfine has 16 bits
       call OAL_obtToTimeTag(obt, time)
       tx(j) = time(0)
       sx(j) = tx(j)
       write(errstr,*)'They map to OBT = ', obt(0), ' and MET = ', time(0)
       call message(AppMsg, Verbose)
    else if(hasAttribute(arr,"DATE-OBS"))then
       tx(j) = totimetag(stringAttribute(arr,"DATE-OBS"))
       sx(j) = tx(j)
       write(errstr,*)'Found DATE-OBS = ', stringAttribute(arr,"DATE-OBS")
       call message(AppMsg, Verbose)
    else
       call setToNull(c_time,j)
    endif
    ix => int8Data(c_ccd)
    if(hasAttribute(arr,"CCDID"))then
       ix(j) = int32attribute(arr,"CCDID")
    else
       call setToNull(c_ccd,j)
    end if
    ix => int8Data(c_node)
    if(inode == 2)then
       call setToNull(c_node,j)    
    else
       ix(j) = inode
    end if
    ix => int8Data(c_ccdocb)
    if(hasAttribute(arr,"CCDOCB"))then
       ix(j) = int32attribute(arr, "CCDOCB")
    else
       call setToNull(c_ccdocb,j)
    end if
    fx => real32Data(c_mean)
    fx(j) = X(1)
    fx => real32Data(c_stddev)
    fx(j) = X(2)
    fx => real32Data(c_norm)
    fx(j) = X(3)
    fx => real32Data(c_euclidenorm)
    fx(j) = euclidenorm
    call release(c_time)
    call release(c_sort)
    call release(c_ccd)
    call release(c_node)
    call release(c_ccdocb)
    call release(c_mean)
    call release(c_stddev)
    call release(c_norm)
    call release(c_euclidenorm)
  end subroutine storeFit

  subroutine storeHisto(histo, set)
    use histogram
    use dal
    type(histogram_t), pointer :: histo
    type(DataSetT), intent(inout) :: set
    type(TableT) :: tbl
    type(ColumnT) :: col
    real(single), pointer :: x(:)
    integer :: i
    integer :: j = 0
    character(len=64) :: cname
    save j
    write(cname,'("phs_",I4.4)')j
    if(j == 0)then
       tbl = addTable(set, "pulse_height_spectra", getNbins(histo))
       col = addColumn(tbl, "pha", real32, "channel")
       x => real32Data(col)
       x = (/ (i, i=0,getNbins(histo)-1) /)
       x = getMinimum(histo) + x * getBinsize(histo)
       call release(col)
    else
       tbl = table(set, "pulse_height_spectra")
    end if
    call addHistory(tbl, "Column dataset mapping: " // trim(cname) // " is " // trim(histo%label))
    col = addColumn(tbl, cname, real32, "counts")
    x => real32Data(col)
    x = getArray(histo)

    call release(col)
    call release(tbl)
    j = j + 1
  end subroutine storeHisto


  ! This subroutine passed to the SLATEC routine DNLS1E
  subroutine gaussian_residuals(IFLAG, M, N, X, FVEC)!, FJAC, LDFJAC)
    !
    ! Given the PHA histogram histo (hx,hy)_i, i=1,...NBINS the
    ! subroutine calculates
    !
    !           FVEC(i) = (hy_i - Gaussian(hx_i,x)
    !
    ! where x contains:
    !   x(1) = mean
    !   x(2) = sigma
    !   x(3) = normalization
    !

    integer, intent(in) :: IFLAG, M, N
    real, dimension(N), intent(in) :: X
    real, dimension(0:M-1), intent(out) :: FVEC

    integer :: i
    real(single), pointer :: histovals(:)
    real(single) :: arg, phaval, phamin, phabin, gest, norm, mean, stddev

    mean = X(1)
    stddev = X(2)
    norm = X(3)

    select case(iflag)
    case(0) ! printing
       write(errstr,*) 'printing not supported'
       call message(appMsg, Verbose)

    case(1) ! calculate function at X and return it in FVEC
       histovals => getArray(histo)
       phamin = getMinimum(histo)
       phabin = getBinsize(histo)
       if(abs(stddev) <= epsilon(1.0))then
          write(errstr,*)'Estimated std dev too small ',stddev
          call warning('residuals')
          stddev = epsilon(1.0)
       end if
       do i=0, M-1
          FVEC(i) = 0
          phaval = phamin + (i + 0.5) * phabin
          arg = (mean - phaval) / stddev
          gest = norm * exp(-arg**2/2)
          FVEC(i) = histovals(i) - gest
          !          if(histovals(i) <= epsilon(1.0))then
          !             histovals(i) = epsilon(1.0)
          !             FVEC(i) = 0
          !          else
          !             FVEC(i) = FVEC(i)/sqrt(histovals(i))
          !          end if
       end do

    case(2:3)
       write(errstr, *) 'Jacobian not calculated'
       call message(appMsg, Verbose)

    case default
       call error("invalidIFLAG", "wrong IFLAG")
    end select


  end subroutine gaussian_residuals

  function getDiagnosticArray(set) result(arr)
    use dal
    type(DataSetT), intent(in) :: set
    type(ArrayT) :: arr

    ! Ugly: the ODF ICD encodes the intrument name and block position
    ! in the block name. Possible names are R1DII1 and R2DII1.
    if(hasBlock(set,"R1DII1"))then
       arr = array(set, "R1DII1", READ)
    elseif(hasBlock(set,"R2DII1"))then
       arr = array(set, "R2DII1", READ)
    else
       write(errstr,*)'Cannot find an array with an ODF compliant name to process in ' //&
            & trim(name(set)) // '. I looked for R1DII1 R2DII1. Trying a default array block.'
       call warning('NoOdfArrayName')
       arr = array(set, trim(name(set)) // ":", READ)
       write(errstr,*)'Trying this block: ' // trim(name(arr))
       call message(appMsg, Sparse)
    end if
  end function getDiagnosticArray


  subroutine initCcfSet(set, calscope)
    use dal
    type(DataSetT), intent(in) :: set
    character(len=*), intent(in) :: calscope
    type(TableT) :: tbl
    type(ColumnT) :: c_time, c_ccd, c_ccdocb, c_mean, c_stddev, c_norm, c_node, c_euclidenorm, c_sort
    
    
    call setAttribute(set, "CATEGORY", "XMMCCF_PIPELINE","")
    call setAttribute(set, "ISSUE", 0, "")
    call setAttribute(set, "VALDATE", "irrelevant", "")
    call setAttribute(set, "SUBDATE", "irrelevant", "")
    call setAttribute(set, "TELESCOP", "XMM", "")
    call setAttribute(set, "INSTRUME", calscope, "")
    call setAttribute(set, "SCOPE", calscope, "")
    call setAttribute(set, "TYPEID", "DIAGNOSTIC_OFFSETS", "")
    call setAttribute(set, "EXTSEQU", "ANY", "")
    call setAttribute(set, "EXTSEQID", "RGSOFFSETS", "")       

    tbl = addTable(set, "RGSOFFSETS", 0)
    c_time = addColumn(tbl, "TIME", real64, "s", (/ 1 /), "from XMM mission reference date")
    c_sort = addColumn(tbl, "TIMESORT", integer32, "s", (/ 1 /), "integer TIME, used to sort the table")
    c_ccd = addColumn(tbl, "CCD_ID", integer8)
    call setNullValue(c_ccd,-1)
    c_node = addColumn(tbl, "NODE_ID", integer8)
    call setNullValue(c_node,-1)
    c_ccdocb = addColumn(tbl, "CCD_OCB", integer8)
    call setNullValue(c_ccdocb,-1)
    c_mean = addColumn(tbl, "MEAN", real32, "channel")
    c_stddev = addColumn(tbl, "STD_DEV", real32, "channel")
    c_norm = addColumn(tbl, "NORMALIZATION", real32, "count")
    c_euclidenorm = addColumn(tbl, "EUCLIDIAN_NORM", real32)

    call release(c_time)
    call release(c_sort)
    call release(c_ccd)
    call release(c_node)
    call release(c_ccdocb)
    call release(c_mean)
    call release(c_stddev)
    call release(c_norm)
    call release(c_euclidenorm)
    
  end subroutine initCcfSet

  function validateImage(set) result(isvalid)
    use dal
    type(DataSetT), intent(in) :: set
    integer(kind=int16), dimension(:,:), pointer :: image
    logical, dimension(:,:), allocatable :: aux
    logical :: isvalid
    type(ArrayT)   :: arr
    integer :: blank, nblanks
    isvalid = .true.
    arr = getDiagnosticArray(set)
    image => int16Array2Data(arr)
    allocate(aux(size(image,1), size(image,2)))
    blank = int32Attribute(arr,"BLANK")
    write(errstr,*)trim(name(set)) // " BLANK = ", blank
    call message(AppMsg, Sparse)
    aux = .true.
    where(image == blank) aux = .false.
    nblanks = size(image) - count(aux)
    if(nblanks/size(image) > nblanks_fraction)then
       write(errstr,*)trim(name(set)) // " contains more than ", nblanks_fraction, " blank pixels."
       call warning("DatasetIgnored")
       write(errstr,*)trim(name(set)) // " ", nblanks, " out of ", size(image), " pixels are blank."
       call message(AppMsg, Sparse)
       isvalid = .false.
    end if
    deallocate(aux)
    return
  end function validateImage
end module rgsoffsetcalc_mod
